import logging
import os
from pathlib import Path
from typing import Dict, final, IO, List, Optional, Set, Tuple, Union

import numpy as np
from pkg_resources import resource_filename
from pymorphy2.tagset import OpencorporaTag

from medspellchecker.tool.abstract_candidate_ranker import AbstractCandidateRanker
from medspellchecker.tool.candidate_generator import CandidateGenerator
from medspellchecker.tool.candidate_word import CandidateWord
from medspellchecker.tool.edit_distance import EditDistanceAlgo
from medspellchecker.tool.mistake_type import MistakeType
from medspellchecker.tool.pre_post_processor import PreProcessor
from medspellchecker.tool.word import Word

logger = logging.getLogger(__name__)


@final
class MedSpellchecker:
    def __init__(self, candidate_ranker: AbstractCandidateRanker,
                 words_list: Optional[Union[Path, str, IO[str], List[str]]] = resource_filename(
                     __name__, "data/processed_lemmatized_all_dict.txt"),
                 encoding: Optional[str] = None,
                 edit_distance_algo: EditDistanceAlgo = EditDistanceAlgo.DAMERAU_OSA_FAST,
                 max_dictionary_edit_distance: int = 1,
                 handle_compound_words: bool = False,
                 saved_state_folder: Optional[Union[Path, str]] = None):
        self._version = 1
        self._handle_compound_words = handle_compound_words
        if isinstance(words_list, (Path, str)):
            corpus = Path(words_list)
            if not corpus.exists():
                logger.error(f"Corpus not found at {corpus}.")
            with open(corpus, "r", encoding=encoding) as infile:
                self.words: Set[str] = set(infile.read().splitlines())
        else:
            self.words: Set[str] = set(words_list)

        self._pre_processor: PreProcessor = PreProcessor()

        if saved_state_folder is not None:
            self._candidate_generator: CandidateGenerator = CandidateGenerator(saved_state_folder=saved_state_folder)
        else:
            self._candidate_generator: CandidateGenerator = CandidateGenerator(
                words_list, encoding, edit_distance_algo, max_dictionary_edit_distance
            )

        self._candidate_ranker: AbstractCandidateRanker = candidate_ranker

    def save_state(self, path: Union[Path, str]):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        self._candidate_generator.save_state(path)

    def _pick_top_missing_space_candidate(self, word: Word,
                                          valid_words_before: List[Word],
                                          valid_words_after: List[Word]) -> Optional[Tuple[CandidateWord, float]]:
        candidate_words = []
        for i in range(len(word.original_value)):
            first_part: Tuple[str, OpencorporaTag] = self._pre_processor.lemmatize_and_tag(word.original_value[:i])
            second_part: Tuple[str, OpencorporaTag] = self._pre_processor.lemmatize_and_tag(word.original_value[i:])
            if first_part[0] in self.words and second_part[0] in self.words:
                candidate_word = f"{first_part[0]} {second_part[0]}"
                first_word: Word = Word(word.position, first_part[0], True, corrected_value=first_part[0])
                second_word: Word = Word(word.position, second_part[0], True, corrected_value=second_part[0])
                first_score = self._candidate_ranker.predict_score(
                    valid_words_before,
                    [second_word] + valid_words_after,
                    first_part[0]
                )
                second_score = self._candidate_ranker.predict_score(
                    valid_words_before + [first_word],
                    valid_words_after,
                    second_part[0]
                )
                if first_score and second_score:
                    candidate_words.append(
                        (CandidateWord(candidate_word, 1),
                         (first_score + second_score) / 2,
                         first_part[1],
                         second_part[1])
                    )
        if candidate_words:
            candidate_words.sort(key=lambda x: x[1])
            best_candidate_word_tuple = candidate_words[0]
            best_candidate_word_split = best_candidate_word_tuple[0].value.split()
            best_candidate_word_part_first: str = best_candidate_word_split[0]
            best_candidate_word_part_second: str = best_candidate_word_split[1]

            restored_first_part: str = self._pre_processor.str_restore_original_form(
                best_candidate_word_part_first, best_candidate_word_tuple[2])
            restored_second_part: str = self._pre_processor.str_restore_original_form(
                best_candidate_word_part_second, best_candidate_word_tuple[3])

            # return CandidateWord(f"{restored_first_part} {restored_second_part}", 1), best_candidate_word_tuple[1]
            return best_candidate_word_tuple[0], best_candidate_word_tuple[1]
        else:
            return None

    # TODO: Do I need to save the formatting after fixing? -- no for now,
    #  usually formatting is not necessary for ml tasks
    def fix_text(self, text: str) -> str:
        # Remove newlines as the MosesTokenizer fails on newlines.
        # So if we decide to keep original text formatting then it should be reworked.
        text_without_newline: str = text.replace("\n", " ")
        # Tokenize text
        tokens: List[str] = self._pre_processor.tokenize(text_without_newline)
        # Build internal representation of words
        words: List[Word] = [word for word in self._pre_processor.generate_words_from_tokens(tokens)]

        current_word_idx = 0
        while current_word_idx < len(words):
            current_word = words[current_word_idx]

            if (not current_word.should_correct) or (current_word.original_value in self.words) or \
                    (current_word.lemma_normal_form in self.words):
                current_word.corrected_value = current_word.original_value
                current_word_idx += 1
                continue

            valid_words_before: List[Word] = [word for word in words[:current_word_idx] if word.should_correct]
            valid_words_after: List[Word] = [word for word in words[current_word_idx + 1:] if word.should_correct]

            mistake_type_to_candidate: Dict[MistakeType, Tuple[CandidateWord, float]] = {}

            # Compute score of original word
            original_word_score: Optional[float] = self._candidate_ranker.predict_score(
                valid_words_before, valid_words_after, current_word.get_lemma_normal_or_original_form())

            if original_word_score is None:
                original_word_score = 1.0

            # If we shouldn't handle spacing mistakes then that's it
            if self._handle_compound_words:
                # Else we need to process handle cases
                # Consider missing space mistake
                if len(current_word.get_lemma_normal_or_original_form()) > 3:
                    top_split_candidate_pair: Optional[Tuple[CandidateWord, float]] = \
                        self._pick_top_missing_space_candidate(current_word, valid_words_before, valid_words_after)
                    if top_split_candidate_pair:
                        mistake_type_to_candidate[MistakeType.MISSING_SPACE_MISTAKE] = top_split_candidate_pair

                # Consider extra space mistake
                compound_word: Optional[Word] = self._create_compound_word(current_word_idx, current_word, words)
                top_extra_space_candidate_pair: Optional[Tuple[CandidateWord, float]] = \
                    self._get_top_extra_space_candidate(compound_word, valid_words_before, valid_words_after)
                if top_extra_space_candidate_pair:
                    mistake_type_to_candidate[MistakeType.EXTRA_SPACE_MISTAKE] = top_extra_space_candidate_pair
            else:
                compound_word = None

            # Consider non-spacing types of mistakes
            # Generate list of candidates for fix
            if len(current_word.get_lemma_normal_or_original_form()) > 3:
                single_candidates_list: List[CandidateWord] = self._candidate_generator.generate_fixing_candidates(
                    current_word, include_unknown=False)

                if self._handle_compound_words and current_word_idx + 1 < len(words):
                    top_single_word_candidate_pair: Optional[Tuple[CandidateWord, float]] = \
                        self._candidate_ranker.pick_top_candidate_with_next_word(
                            current_word,
                            words[current_word_idx + 1].get_lemma_normal_or_original_form(),
                            valid_words_before,
                            valid_words_after,
                            single_candidates_list
                        )
                else:
                    top_single_word_candidate_pair: Optional[Tuple[CandidateWord, float]] = \
                        self._candidate_ranker.pick_top_candidate(current_word, valid_words_before,
                                                                  valid_words_after, single_candidates_list)
                if top_single_word_candidate_pair:
                    mistake_type_to_candidate[MistakeType.SINGLE_WORD_MISTAKE] = top_single_word_candidate_pair

            # Early skip
            if not mistake_type_to_candidate:
                current_word.corrected_value = current_word.get_lemma_normal_or_original_form()
                current_word_idx += 1
                continue

            # Now we have collected all possible candidates. So let's pick most suitable one
            top_candidate: Tuple[MistakeType, Tuple[CandidateWord, float]] = \
                self._find_top_candidate(mistake_type_to_candidate)
            self._fix_word_by_top_candidate(
                top_candidate,
                original_word_score,
                current_word,
                current_word_idx,
                words,
                compound_word
            )
            current_word_idx += 1

        corrected_text = self._build_result_text(words)
        return corrected_text

    def _create_compound_word(self, current_word_idx: int, current_word: Word, words: List[Word]) -> Optional[Word]:
        if current_word_idx + 1 < len(words):
            next_word: Word = words[current_word_idx + 1]
            new_word_str: str = current_word.original_value + next_word.original_value
            new_word: Word = next(self._pre_processor.generate_words_from_tokens([new_word_str]))
            return new_word
        return None

    def _get_top_extra_space_candidate(self, compound_word: Optional[Word],
                                       valid_words_before: List[Word],
                                       valid_words_after: List[Word]) -> Optional[Tuple[CandidateWord, float]]:
        if compound_word and compound_word.should_correct and compound_word.get_lemma_normal_or_original_form() in self.words:
            extra_space_candidate: CandidateWord = CandidateWord(compound_word.get_lemma_normal_or_original_form(), 1)
            top_extra_space_candidate_pair: Optional[Tuple[CandidateWord, float]] = \
                self._candidate_ranker.pick_top_candidate(
                    compound_word, valid_words_before, valid_words_after[1:], [extra_space_candidate])
            return top_extra_space_candidate_pair
        return None

    def _fix_word_by_top_candidate(self,
                                   top_candidate: Tuple[MistakeType, Tuple[CandidateWord, float]],
                                   original_word_score: float,
                                   current_word: Word,
                                   current_word_idx: int,
                                   words: List[Word],
                                   compound_word: Optional[Word]):
        top_candidate_type: MistakeType = top_candidate[0]
        top_candidate_word: CandidateWord = top_candidate[1][0]
        top_candidate_score: float = top_candidate[1][1]

        if original_word_score < top_candidate_score:
            if top_candidate_type == MistakeType.SINGLE_WORD_MISTAKE:
                current_word.corrected_value = top_candidate_word.value
                return
            elif self._handle_compound_words:
                if top_candidate_type == MistakeType.MISSING_SPACE_MISTAKE:
                    # Here we didn't create two separate words due to the performance reasons
                    current_word.corrected_value = top_candidate_word.value
                    return
                elif top_candidate_type == MistakeType.EXTRA_SPACE_MISTAKE:
                    del words[current_word_idx + 1]
                    compound_word.corrected_value = top_candidate_word.value
                    words[current_word_idx] = compound_word
                    return

        current_word.corrected_value = current_word.get_lemma_normal_or_original_form()

    def _find_top_candidate(self, mistake_type_to_candidate: Dict[MistakeType, Tuple[CandidateWord, float]]) \
            -> Tuple[MistakeType, Tuple[CandidateWord, float]]:
        mistake_type_candidate_list: List[Tuple[MistakeType, Tuple[CandidateWord, float]]] = \
            list(mistake_type_to_candidate.items())

        candidate_scores: List[float] = list(map(lambda x: x[1][1], mistake_type_candidate_list))
        top_candidate_idx: int = np.argmax(candidate_scores)
        return mistake_type_candidate_list[top_candidate_idx]

    def _build_result_text(self, words: List[Word]) -> str:
        result_list = [self._pre_processor.word_restore_original_form(word) for word in words]
        return ' '.join(result_list)



from abc import abstractmethod
import dataclasses
import itertools
import logging
from typing import List, Optional, Tuple

from accelerate import Accelerator
import torch
from torch import IntTensor
from transformers import AutoModelForMaskedLM, AutoTokenizer, PreTrainedModel, PreTrainedTokenizer

from medspellchecker.tool.abstract_candidate_ranker import AbstractCandidateRanker
from medspellchecker.tool.candidate_word import CandidateWord
from medspellchecker.tool.gpu_utils import set_device
from medspellchecker.tool.word import Word

logger = logging.getLogger(__name__)


class AbstractBertCandidateRanker(AbstractCandidateRanker):

    def __init__(self, pretrained_model_checkpoint: str,
                 use_treshold: bool = True, treshold: float = 0.0000001,
                 use_gpu: bool = True):
        self._use_treshold: bool = use_treshold
        self._treshold: float = treshold
        self._use_gpu: bool = set_device() if use_gpu else False

        accelerator: Accelerator = Accelerator(cpu=not self._use_gpu)
        self._tokenizer: PreTrainedTokenizer = accelerator.prepare(
            AutoTokenizer.from_pretrained(pretrained_model_checkpoint))
        self._model: PreTrainedModel = accelerator.prepare(
            AutoModelForMaskedLM.from_pretrained(pretrained_model_checkpoint))

    @property
    @abstractmethod
    def _version(self) -> int:
        """
        :return: version of ranker
        """

    def prepare_text_for_prediction(self, correct_words_before: List[Word],
                                    words_after: List[Word]) -> str:
        str_before = ' '.join([word.corrected_value for word in correct_words_before])
        str_word_token = self._tokenizer.mask_token
        str_after = ' '.join([word.original_value for word in words_after])

        return ' '.join([str_before, str_word_token, str_after])

    def create_texts_for_prediction_with_ids(self, input_ids: List[int],
                                             candidate_ids: List[int]) -> List[Tuple[IntTensor, IntTensor]]:
        results: List[Tuple[IntTensor, IntTensor]] = []
        input_ids_mask_pos: int = input_ids.index(self._tokenizer.mask_token_id)
        input_ids_before: List[int] = input_ids[:input_ids_mask_pos]
        input_ids_after: List[int] = input_ids[input_ids_mask_pos + 1:]
        for i in range(len(candidate_ids)):
            candidate_ids_before: List[int] = candidate_ids[:i]
            candidate_ids_after: List[int] = candidate_ids[i + 1:]
            patched_input_ids: IntTensor = IntTensor([list(itertools.chain.from_iterable(
                [input_ids_before, candidate_ids_before,
                 [self._tokenizer.mask_token_id], candidate_ids_after, input_ids_after]))])
            patched_attention_mask: IntTensor = IntTensor([list(itertools.repeat(1, patched_input_ids.size()[1]))])
            results.append((patched_input_ids, patched_attention_mask))
        return results

    def predict_score(self, correct_words_before: List[Word], words_after: List[Word], candidate_value: str) -> \
            Optional[float]:
        text_with_mask = self.prepare_text_for_prediction(correct_words_before, words_after)
        candidate_token_ids = self._tokenizer.encode(candidate_value, add_special_tokens=False)
        logger.debug(f"Candidate token ids: {candidate_token_ids}")

        inputs = self._tokenizer(text_with_mask, return_tensors='pt')
        logger.debug(f"Input token ids: {inputs}")

        original_input_ids = inputs["input_ids"][0].tolist()
        patched_inputs_data = self.create_texts_for_prediction_with_ids(original_input_ids, candidate_token_ids)

        ## Mean score of candidate word parts
        candidate_mean_score = 0.0
        for i, patched_item in enumerate(patched_inputs_data):
            patched_inputs_ids, patched_attention_mask = patched_item
            if self._use_gpu:
                patched_inputs_ids = patched_inputs_ids.cuda()
                patched_attention_mask = patched_attention_mask.cuda()

            token_logits = self._model(patched_inputs_ids, patched_attention_mask).logits
            logger.debug(f"Token logits: {token_logits}")

            mask_token_index = torch.where(inputs["input_ids"] == self._tokenizer.mask_token_id)[1]
            mask_token_logits = token_logits[0, mask_token_index, :]
            softmax_mask_token_logits = torch.softmax(mask_token_logits, dim=1)

            if logger.getEffectiveLevel() == logging.DEBUG:
                top_20 = torch.topk(softmax_mask_token_logits, 20, dim=1)
                top_20_tokens = zip(top_20.indices[0].tolist(), top_20.values[0].tolist())
                logger.debug("----Top 20 most suitable tokens----")
                for token, score in top_20_tokens:
                    logger.debug(f"{self._tokenizer.decode([token])}, score: {score}")

            candidate_mean_score += softmax_mask_token_logits[:, candidate_token_ids[i]][0].item()
        candidate_mean_score = candidate_mean_score / len(candidate_token_ids)
        logger.debug(f"Candidate - {candidate_value}, score {candidate_mean_score}")
        if self._use_treshold and candidate_mean_score > self._treshold or not self._use_treshold:
            return candidate_mean_score
        else:
            return None

    def rank_candidates(self, current_word: Word,
                        correct_words_before: List[Word],
                        words_after: List[Word],
                        candidates: List[CandidateWord]) -> List[CandidateWord]:
        if not candidates:
            return candidates

        # candidates must be ordered by edit distance
        candidates.sort(key=lambda x: x.distance)

        candidate_score_pairs = []

        for candidate in candidates:
            candidate_score = self.predict_score(correct_words_before, words_after, candidate.value)
            if candidate_score:
                candidate_score_pairs.append((candidate, candidate_score))

        ranked_candidate_score_pairs = sorted(candidate_score_pairs, key=lambda x: x[1])
        ranked_candidates = map(lambda x: x[0], ranked_candidate_score_pairs)
        return list(ranked_candidates)

    def pick_top_candidate(self, current_word: Word,
                           correct_words_before: List[Word],
                           words_after: List[Word],
                           candidates: List[CandidateWord]) -> Optional[Tuple[CandidateWord, float]]:
        if not candidates:
            return None

        # candidates must be ordered by edit distance
        candidates.sort(key=lambda x: x.distance)

        candidate_score_pairs = []
        candidates = list(filter(lambda x: x.distance == candidates[0].distance, candidates))

        for candidate in candidates:
            candidate_score = self.predict_score(correct_words_before, words_after, candidate.value)
            if candidate_score:
                candidate_score_pairs.append((candidate, candidate_score))

        if candidate_score_pairs:
            ranked_candidate_score_pairs = sorted(candidate_score_pairs, key=lambda x: x[1])
            return ranked_candidate_score_pairs[0]
        else:
            return None

    def pick_top_candidate_with_next_word(self, current_word: Word,
                                          next_word_str: str,
                                          correct_words_before: List[Word],
                                          words_after: List[Word],
                                          candidates: List[CandidateWord]) -> Optional[Tuple[CandidateWord, float]]:
        if not candidates:
            return None

        # candidates must be ordered by edit distance
        candidates.sort(key=lambda x: x.distance)

        candidate_score_pairs = []
        candidates = list(filter(lambda x: x.distance == candidates[0].distance, candidates))

        for candidate in candidates:
            candidate_score = self.predict_score(correct_words_before, words_after, candidate.value)
            corrected_current_word = dataclasses.replace(current_word)
            corrected_current_word.corrected_value = candidate.value
            before = correct_words_before + [corrected_current_word]
            after = words_after[1:]
            next_word_score = self.predict_score(before, after, next_word_str)
            if candidate_score and next_word_score:
                candidate_score_pairs.append((candidate, (candidate_score + next_word_score) / 2))

        if candidate_score_pairs:
            ranked_candidate_score_pairs = sorted(candidate_score_pairs, key=lambda x: x[1])
            return ranked_candidate_score_pairs[0]
        else:
            return None



import logging
import subprocess

import numpy as np
import pynvml
import torch

# 8 Gb
MINIMAL_REQUIRED_GRU_MEMORY = 8192


def set_device() -> bool:
    """Set the most appropriate device for torch

       Returns:
            True if CUDA device selected, otherwise False
   """

    def set_cpu_device():
        torch.device("cpu")
        logging.info(f"We will use device: CPU")

    try:
        nvidia_output = subprocess.check_output("nvidia-smi", shell=True)
        logging.info(nvidia_output)
    except subprocess.CalledProcessError:
        logging.info('Nvidia driver not installed')
        set_cpu_device()
        return False

    pynvml.nvmlInit()

    if torch.cuda.is_available():
        gpus_free_mem_list = []
        for device_num in range(pynvml.nvmlDeviceGetCount()):
            handle = pynvml.nvmlDeviceGetHandleByIndex(device_num)
            info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            gpus_free_mem_list.append((info.total - info.used) // 1024 ** 3)
        selected_gpu_device_number = np.argmax(gpus_free_mem_list)
        handle = pynvml.nvmlDeviceGetHandleByIndex(selected_gpu_device_number)
        info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        selected_gpu_device_memory = (info.total - info.used) // 1024 ** 2

        if selected_gpu_device_memory < MINIMAL_REQUIRED_GRU_MEMORY:
            logging.info("All available GPUs are busy and don't have required free memory")
            set_cpu_device()
            return False

        torch.cuda.set_device(torch.device(selected_gpu_device_number))
        logging.info(f"Selected GPU number: {torch.cuda.current_device()}")
        logging.info(f"Will use device {torch.cuda.current_device()}: "
                     f"{torch.cuda.get_device_name(torch.cuda.current_device())}")
        logging.info(f"Device has {selected_gpu_device_memory} Mb free memory")
        return True
    else:
        logging.info("There is no available GPU")
        set_cpu_device()
        return False


def print_gpu_memory_stats():
    current_device = torch.cuda.current_device()
    handle = pynvml.nvmlDeviceGetHandleByIndex(current_device)
    info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    logging.info(f"All GPU memory occupied: {info.used // 1024 ** 3}/{info.total // 1024 ** 3}  Gb.")
    logging.info(
        f"Torch GPU {current_device} memory allocated: {torch.cuda.memory_allocated(current_device) // 1024 ** 3} Gb")



import logging
from typing import final

from medspellchecker.tool.abstract_bert_candidate_ranker import AbstractBertCandidateRanker

logger = logging.getLogger(__name__)


@final
class RuDistilBertCandidateRanker(AbstractBertCandidateRanker):
    _pretrained_model_checkpoint: str = "DmitryPogrebnoy/MedDistilBertBaseRuCased"

    def __init__(self, use_treshold: bool = True, use_gpu: bool = True):
        super().__init__(RuDistilBertCandidateRanker._pretrained_model_checkpoint, use_treshold, 0.0000001, use_gpu)

    @property
    def _version(self) -> int:
        return 1



from re import search
from typing import final, Final, Generator, List, Optional, Tuple

from nltk import download
from nltk.corpus import stopwords
from pymorphy2 import MorphAnalyzer
from pymorphy2.analyzer import Parse
from pymorphy2.shapes import restore_capitalization
from pymorphy2.tagset import OpencorporaTag
from sacremoses import MosesTokenizer

from medspellchecker.tool.word import Word


@final
class PreProcessor:
    _stopwords_download_name: Final[str] = "stopwords"

    def __init__(self):
        self._version = 1
        self._tokenizer: Final[MosesTokenizer] = MosesTokenizer(lang="ru")
        self._lemmatizer: Final[MorphAnalyzer] = MorphAnalyzer()
        # Download nltk stopwords dict
        download(PreProcessor._stopwords_download_name, quiet=True)
        self._stopwords: Final[List[str]] = stopwords.words('russian')

    def is_valid_token(self, token: str) -> bool:
        """Checks if the correction token is valid.

        The token must not contain punctuation marks or other symbols,
        otherwise we cannot fix it correctly. And also all the letters together in the token are not capitalized -
        this is to exclude the correction of various abbreviations that are written with capital letters.
        """
        return (not search("[^а-яА-Я]", token)) and (not token.isupper()) and (
            not token in self._stopwords)

    def _is_proper_name(self, token: str, token_id: int, tokens: List[str]) -> bool:
        if token_id == 0:
            return False

        previous_token_id: int = token_id - 1
        previous_token: str = tokens[previous_token_id]

        return (not previous_token.endswith(".")) and token[0].isupper()

    def tokenize(self, string: str) -> List[str]:
        return self._tokenizer.tokenize(string)

    def lemmatize(self, string: str) -> str:
        return self._lemmatizer.parse(string)[0].normal_form

    def lemmatize_and_tag(self, string: str) -> Tuple[str, OpencorporaTag]:
        parse = self._lemmatizer.parse(string)[0]
        return (parse.normal_form, parse.tag)

    def generate_words_from_tokens(self, tokens: List[str]) -> Generator[Word, None, None]:
        for id, token in enumerate(tokens):
            is_valid = self.is_valid_token(token)
            is_proper_name = self._is_proper_name(token, id, tokens)
            if is_valid and not is_proper_name:
                # if the token is valid for correction, extract a lemma from it
                parse: Parse = self._lemmatizer.parse(token)[0]
                if parse.is_known:
                    yield Word(id, token, True, parse.normal_form, parse.tag, parse.lexeme)
                else:
                    yield Word(id, token, True)
            else:
                # else just set corrected_value as original token
                yield Word(id, token, False, corrected_value=token)

    def str_restore_original_form(self, word: str, tag: OpencorporaTag):
        result_lexeme: Parse = self._restore_original_form(word, tag)

        if result_lexeme:
            return result_lexeme.word
        else:
            return word

    def word_restore_original_form(self, word: Word) -> Optional[str]:
        if (not word.should_correct) or (word.lexeme is None) or (word.lemma_tag is None):
            return word.corrected_value

        filtered_lexeme = self._restore_original_form(word.corrected_value, word.lemma_tag)

        if filtered_lexeme:
            return restore_capitalization(filtered_lexeme.word, word.original_value)
        else:
            return restore_capitalization(word.corrected_value, word.original_value)

    def _restore_original_form(self, corrected_word: str, tag: OpencorporaTag) -> Optional[Parse]:
        corrected_word_lexeme = self._lemmatizer.parse(corrected_word)[0].lexeme
        filtered_lexeme = list(filter(
            lambda parse: (parse.tag == tag) and (parse.word[0] == corrected_word[0]),
            corrected_word_lexeme)
        )
        if filtered_lexeme:
            return filtered_lexeme[0]
        else:
            return None



from dataclasses import dataclass
from typing import final, List, Optional

from pymorphy2.analyzer import Parse
from pymorphy2.tagset import OpencorporaTag


@final
@dataclass
class Word:
    """Input word representation.

        Args:
            source_value: The original word.
            should_correct: Should this word be corrected?
            lemma_normal_form: The normal form of the word
            lemma_tag: Tag line of the original word form
            corrected_value: Final corrected word
    """
    position: int
    original_value: str
    should_correct: bool
    lemma_normal_form: Optional[str] = None
    lemma_tag: Optional[OpencorporaTag] = None
    lexeme: Optional[List[Parse]] = None
    corrected_value: Optional[str] = None

    def get_lemma_normal_or_original_form(self):
        if self.lemma_normal_form:
            return self.lemma_normal_form
        else:
            return self.original_value



import json
import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import final, Optional, IO, List, Union, Final, Set, Dict

from medspellchecker.tool.candidate_word import CandidateWord
from medspellchecker.tool.edit_distance import EditDistanceAlgo, EditDistance
from medspellchecker.tool.word import Word

logger = logging.getLogger(__name__)


@final
class CandidateGenerator:
    _saved_state_file_candidate_generator = "candidate_generator_state.json"

    def __init__(self,
                 words_list: Optional[Union[Path, str, IO[str], List[str]]] = None,
                 encoding: Optional[str] = None,
                 edit_distance_algo: EditDistanceAlgo = EditDistanceAlgo.DAMERAU_OSA_FAST,
                 max_dictionary_edit_distance: int = 2,
                 saved_state_folder: Optional[Union[Path, str]] = None):
        self._version = 1

        if saved_state_folder is not None:
            self._load_sate(saved_state_folder)
        else:
            self._edit_distance_comparer: EditDistance = EditDistance(edit_distance_algo)
            self._max_dictionary_edit_distance: Final[int] = max_dictionary_edit_distance

            self._max_dictionary_word_length: int = 0
            self._words_dictionary: Set[str] = set()
            self._deletes_word_dictionary: Dict[str, List[str]] = defaultdict(list)
            # Fill _words_dictionary and build _deletes_word_dictionary
            if words_list is None:
                raise ValueError(f"word_list and saved_state cannot be None at the same time. Pass one of these "
                                 f"arguments!")
            self._create_dictionary_from_words(words_list, encoding)

    def _load_sate(self, saved_state_folder: Union[Path, str]):
        saved_state_file = os.path.join(saved_state_folder, CandidateGenerator._saved_state_file_candidate_generator)

        state_path = Path(saved_state_file)
        if not state_path.exists():
            raise ValueError(f"Load CandidateGenerator state not found at {state_path}.")
        with open(state_path, "r") as infile:
            data = json.load(infile)

        if data["_version"] != self._version:
            raise ValueError(f"Incompatible version of loaded state! Loaded version is {data['_version']}, but "
                             f"spellchecker version is {self._version}")

        self._edit_distance_comparer = EditDistance(data["_edit_distance_algo"])
        # noinspection PyFinal
        self._max_dictionary_edit_distance = data["_max_dictionary_edit_distance"]
        self._max_dictionary_word_length = data["_max_dictionary_word_length"]
        self._words_dictionary = set(data["_words_dictionary"])
        self._deletes_word_dictionary = data["_deletes_word_dictionary"]

    def save_state(self, path: Union[Path, str]):
        data = {
            "_version": self._version,
            "_edit_distance_algo": self._edit_distance_comparer.algorithm,
            "_max_dictionary_edit_distance": self._max_dictionary_edit_distance,
            "_max_dictionary_word_length": self._max_dictionary_word_length,
            "_words_dictionary": list(self._words_dictionary),
            "_deletes_word_dictionary": self._deletes_word_dictionary
        }
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(Path(os.path.join(path, CandidateGenerator._saved_state_file_candidate_generator)), "w") as file:
            file.write(json.dumps(data))

    def _create_dictionary_from_words(
            self,
            words_list: Union[Path, str, IO[str], List[str]],
            encoding: Optional[str] = None) -> bool:
        """Create dictionary words from a file containing words. One word per line.

        Args:
            words_list: The path+filename of the file or afile object of the
                word-list or list of strings.
            encoding: Text encoding of the corpus file.

        Returns:
            ``True`` if file loaded, or ``False`` if file not found.
        """
        if isinstance(words_list, (Path, str)):
            corpus = Path(words_list)
            if not corpus.exists():
                logger.error(f"Corpus not found at {corpus}.")
                return False
            with open(corpus, "r", encoding=encoding) as infile:
                for line in infile.read().splitlines():
                    self._create_dictionary_entry(line)
        else:
            for line in words_list:
                self._create_dictionary_entry(line)
        return True

    def _create_dictionary_entry(self, key: str) -> None:
        """Creates/updates an entry in the dictionary.

        For every word there are deletes with an edit distance of
        1..max_edit_distance created and added to the dictionary. Every delete
        entry has a suggestions list, which points to the original term(s) it was
        created from. The dictionary may be dynamically updated (word frequency
        and new words) at any time by calling create_dictionary_entry.

        Args:
            key: The word to add to dictionary.

        Returns:
            None
        """
        # TODO: Should we lemmatise words in order to glue together several different forms
        #  of the same word and thereby reduce the size of the dictionary?
        if key not in self._words_dictionary:
            self._words_dictionary.add(key)

            # edits/suggestions are created only once, no matter how often word
            # occurs. edits/suggestions are created as soon as the word occurs in the
            # corpus, even if the same term existed before in the dictionary as an
            # edit from another word
            if len(key) > self._max_dictionary_word_length:
                self._max_dictionary_word_length = len(key)

            # create deletes
            edits = self._generate_delete_words(key)
            logger.info(edits)
            for delete in edits:
                self._deletes_word_dictionary[delete].append(key)
        return

    def _recursive_generate_delete_words(
            self,
            word: str,
            edit_distance: int,
            delete_words: Set[str],
            current_distance: int = 0,
    ) -> Set[str]:
        """SymDel algorithm in the recursion style uses only deletes for simulate
        the four expensive operations (transposes, replaces, deletions, addition).
        """
        edit_distance += 1
        if not word:
            return delete_words
        for i in range(current_distance, len(word)):
            delete = word[:i] + word[i + 1:]
            if delete in delete_words:
                continue
            delete_words.add(delete)
            # start recursion, if maximum edit distance not yet reached
            if edit_distance < self._max_dictionary_edit_distance:
                self._recursive_generate_delete_words(delete, edit_distance, delete_words, i)
        return delete_words

    def _generate_delete_words(self, word: str) -> Set[str]:
        hash_set = {word}
        if len(word) <= self._max_dictionary_edit_distance:
            hash_set.add("")
        return self._recursive_generate_delete_words(word, 0, hash_set)

    def generate_fixing_candidates(
            self,
            word: Word,
            max_edit_distance: Optional[int] = None,
            include_unknown: bool = True) -> List[CandidateWord]:

        if max_edit_distance is None:
            max_edit_distance = self._max_dictionary_edit_distance
        if max_edit_distance > self._max_dictionary_edit_distance:
            raise ValueError("Passed edit distance too big!")
        suggestions: List[CandidateWord] = list()

        original_word = word.original_value
        original_word_length = len(original_word)

        def early_return():
            if include_unknown and not suggestions:
                suggestions.append(CandidateWord(original_word, 0))
            return suggestions

        # early return - word is too big to possibly match any words
        if original_word_length - max_edit_distance > self._max_dictionary_word_length:
            return early_return()

        # early_return - quick check for exact match
        if original_word in self._words_dictionary:
            # return exact match
            return early_return()

        # early return - if we only want to check if word in dictionary
        if max_edit_distance == 0:
            return early_return()

        considered_deletes: Set[str] = set()
        considered_candidates: Set[str] = set()
        # we considered the original word already in the 'original_word in self._words_dictionary' above
        considered_candidates.add(original_word)

        candidate_pointer: int = 0
        candidates: List[str] = [original_word]

        while candidate_pointer < len(candidates):
            candidate = candidates[candidate_pointer]
            candidate_pointer += 1
            candidate_len = len(candidate)
            len_diff = abs(original_word_length - candidate_len)

            # early return: if candidate distance is already higher than
            # suggestion distance, then there are no better suggestions to be expected
            if len_diff > max_edit_distance:
                break  # "peephole" optimization, http://bugs.python.org/issue2506

            if candidate in self._deletes_word_dictionary:
                candidates_from_delete_dictionary = self._deletes_word_dictionary[candidate]
                for suggestion in candidates_from_delete_dictionary:
                    if suggestion == original_word:
                        continue
                    suggestion_len = len(suggestion)
                    # original word length and suggestion lengths diff > allowed/current best
                    # distance
                    if (
                            abs(suggestion_len - original_word_length) > max_edit_distance
                            # suggestion must be for a different delete string, in
                            # same bin only because of hash collision
                            or suggestion_len < candidate_len
                            # if suggestion len = delete len, then it either equals
                            # delete or is in same bin only because of hash collision
                            or (suggestion_len == candidate_len and suggestion != candidate)):
                        continue  # "peephole" optimization, http://bugs.python.org/issue2506

                    # True Damerau-Levenshtein Edit Distance: adjust distance,
                    # if both distances>0. We allow simultaneous edits (deletes)
                    # of max_edit_distance on both the dictionary and the
                    # original term. For replaces and adjacent transposes the
                    # resulting edit distance stays <= max_edit_distance. For
                    # inserts and deletes the resulting edit distance might
                    # exceed max_edit_distance. To prevent suggestions of a
                    # higher edit distance, we need to calculate the resulting
                    # edit distance, if there are simultaneous edits on both
                    # sides. Example: (bank==bnak and bank==bink, but bank!=kanb
                    # and bank!=xban and bank!=baxn for max_edit_distance=1).
                    # Two deletes on each side of a pair makes them all equal,
                    # but the first two pairs have edit distance=1, the others
                    # edit distance=2.
                    distance = 0
                    min_distance = 0
                    if candidate_len == 0:
                        # suggestions which have no common chars with phrase
                        # (phrase_len<=max_edit_distance && suggestion_len<=max_edit_distance)
                        distance = max(original_word_length, suggestion_len)
                        if distance > max_edit_distance or suggestion in considered_candidates:
                            continue
                    elif suggestion_len == 1:
                        # This should always be phrase_len - 1? Since
                        # suggestions are generated from deletes of the input
                        # phrase
                        distance = (
                            original_word_length
                            if original_word.index(suggestion[0]) < 0
                            else original_word_length - 1
                        )
                        # `suggestion` only gets added to
                        # `considered_candidates` when `suggestion_len>1`.
                        # Given the max_dictionary_edit_distance and
                        # prefix_length restrictions, `distance`` should never
                        # be >max_edit_distance_2
                        if (
                                distance > max_edit_distance
                                or suggestion in considered_candidates
                        ):
                            continue
                    # number of edits in prefix ==maxeditdistance AND no
                    # identical suffix, then editdistance>max_edit_distance and
                    # no need for Levenshtein calculation
                    # (phraseLen >= prefixLength) &&
                    # (suggestionLen >= prefixLength)
                    else:
                        # delete_in_suggestion_prefix is somewhat expensive
                        if suggestion in considered_candidates:
                            continue
                        considered_candidates.add(suggestion)
                        distance = self._edit_distance_comparer.compare(
                            original_word, suggestion, max_edit_distance
                        )
                        if distance < 0:
                            continue
                    # do not process higher distances than those already found,
                    # if verbosity<ALL (note: max_edit_distance_2 will always
                    # equal max_edit_distance when Verbosity.ALL)
                    if distance <= max_edit_distance:  # pragma: no branch
                        item = CandidateWord(suggestion, distance)
                        suggestions.append(item)

            # add edits: derive edits (deletes) from candidate (phrase) and add
            # them to candidates list. this is a recursive process until the
            # maximum edit distance has been reached
            if len_diff < max_edit_distance:
                for i in range(candidate_len):
                    delete = candidate[:i] + candidate[i + 1:]
                    if delete not in considered_deletes:
                        considered_deletes.add(delete)
                        candidates.append(delete)

        early_return()
        return suggestions



import pathlib

from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

print(find_packages(where='spellchecker'))

setup(
    name='medspellchecker',
    version='0.0.2',
    description='Fast and effective spellchecker for Russian medical texts',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DmitryPogrebnoy/MedSpellChecker',
    author='Dmitry Pogrebnoy',
    author_email='pogrebnoy.inc@gmail.com',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing',
        'Topic :: Text Editors :: Word Processors',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Environment :: GPU :: NVIDIA CUDA',
    ],
    keywords='spellchecker, nlp, medical, text correction',
    package_dir={'medspellchecker.tool': 'medspellchecker/tool'},
    packages=['medspellchecker.tool'],
    python_requires='>=3.9',
    install_requires=['pymorphy2>=0.9.1',
                      'torch>=1.11.0',
                      'numpy>=1.22',
                      'scipy>=1.7.3',
                      'tqdm>=4.62.3',
                      'nltk>=3.6.7',
                      'editdistpy>=0.1.3',
                      'accelerate>=0.12.0',
                      'transformers>=4.22.2',
                      'sacremoses>=0.0.49',
                      'pynvml>=11.4.1'],
    include_package_data=True,
    package_data={'medspellchecker.tool': ['data/processed_lemmatized_all_dict.txt']},
    project_urls={
        'Bug Reports': 'https://github.com/DmitryPogrebnoy/MedSpellChecker/issues',
        'Funding': 'https://donate.pypi.org',
    },
)



from dataclasses import dataclass
from typing import final


@final
@dataclass
class CandidateWord:
    """Spelling suggestion.

        Args:
            value: The suggested word.
            distance: Edit distance from search word.
    """

    value: str
    distance: int
    # TODO: Add more features for ranking if needed



from abc import ABC, abstractmethod
from enum import IntEnum

from editdistpy import damerau_osa, levenshtein


class EditDistanceAlgo(IntEnum):
    """Supported edit distance algorithms."""

    LEVENSHTEIN_FAST = 0  #: Fast Levenshtein algorithm.
    DAMERAU_OSA_FAST = 1  #: Fast Damerau optimal string alignment algorithm


class EditDistance:
    """Edit distance algorithms.

    Args:
        algorithm: The distance algorithm to use.

    Attributes:
        _algorithm (:class:`EditDistanceAlgo`): The edit distance algorithm to
            use.
        _distance_comparer (:class:`AbstractDistanceComparer`): An object to
            compute the relative distance between two strings. The concrete
            object will be chosen based on the value of :attr:`_algorithm`.

    Raises:
        ValueError: If `algorithm` specifies an invalid distance algorithm.
    """

    def __init__(self, algorithm: EditDistanceAlgo) -> None:
        self._distance_comparer: AbstractDistanceComparer
        self.algorithm = algorithm
        if algorithm == EditDistanceAlgo.LEVENSHTEIN_FAST:
            self._distance_comparer = LevenshteinFast()
        elif algorithm == EditDistanceAlgo.DAMERAU_OSA_FAST:
            self._distance_comparer = DamerauOsaFast()
        else:
            raise ValueError("Unknown distance algorithm!")

    def compare(self, string_1: str, string_2: str, max_distance: int) -> int:
        """Compares a string to the base string to determine the edit distance,
        using the previously selected algorithm.

        Args:
            string_1: Base string.
            string_2: The string to compare.
            max_distance: The maximum distance allowed.

        Returns:
            The edit distance (or -1 if `max_distance` exceeded).
        """
        return self._distance_comparer.distance(string_1, string_2, max_distance)


class AbstractDistanceComparer(ABC):
    """An interface to compute relative distance between two strings."""

    @abstractmethod
    def distance(self, string_1: str, string_2: str, max_distance: int) -> int:
        """Returns a measure of the distance between two strings.

        Args:
            string_1: One of the strings to compare.
            string_2: The other string to compare.
            max_distance: The maximum distance that is of interest.

        Returns:
            -1 if the distance is greater than the max_distance, 0 if the strings
                are equivalent, otherwise a positive number whose magnitude
                increases as difference between the strings increases.
        """


class LevenshteinFast(AbstractDistanceComparer):
    """Provides an interface for computing edit distance metric between two
    strings using the fast Levenshtein algorithm.
    """

    def distance(self, string_1: str, string_2: str, max_distance: int) -> int:
        """Computes the Levenshtein edit distance between two strings.

        Args:
            string_1: One of the strings to compare.
            string_2: The other string to compare.
            max_distance: The maximum distance that is of interest.

        Returns:
            -1 if the distance is greater than the max_distance, 0 if the strings
                are equivalent, otherwise a positive number whose magnitude
                increases as difference between the strings increases.
        """
        return levenshtein.distance(string_1, string_2, max_distance)


class DamerauOsaFast(AbstractDistanceComparer):
    """Provides an interface for computing edit distance metric between two
    strings using the fast Damerau-Levenshtein Optimal String Alignment (OSA)
    algorithm.
    """

    def distance(self, string_1: str, string_2: str, max_distance: int) -> int:
        """Computes the Damerau-Levenshtein optimal string alignment edit
        distance between two strings.

        Args:
            string_1: One of the strings to compare.
            string_2: The other string to compare.
            max_distance: The maximum distance that is of interest.

        Returns:
            -1 if the distance is greater than the max_distance, 0 if the strings
                are equivalent, otherwise a positive number whose magnitude
                increases as difference between the strings increases.
        """
        return damerau_osa.distance(string_1, string_2, max_distance)



from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from medspellchecker.tool.candidate_word import CandidateWord
from medspellchecker.tool.word import Word


class AbstractCandidateRanker(ABC):
    @abstractmethod
    def predict_score(self, correct_words_before: List[Word], words_after: List[Word], candidate_value: str) -> \
    Optional[float]:
        """Returns score of candidate word

            Args:
                current_word: current word
                correct_words_before: corrected words before
                words_after: words after (not corrected yet)
                candidate_value: Value for computing score

            Returns:
                Score of candidate word or None if it is under treshold
        """

    @abstractmethod
    def rank_candidates(self, current_word: Word,
                        correct_words_before: List[Word],
                        words_after: List[Word],
                        candidates: List[CandidateWord]) -> List[CandidateWord]:
        """Returns ranked candidate words for correction

        Args:
            current_word: The words that need correction
            correct_words_before: correct words before current word
            words_after: not processed yet words after current word
            candidates: Candidates for right word

        Returns:
            Ranked candidates. First element is the most suitable for fixing words
        """

    @abstractmethod
    def pick_top_candidate(self, current_word: Word,
                           correct_words_before: List[Word],
                           words_after: List[Word],
                           candidates: List[CandidateWord]) -> Optional[Tuple[CandidateWord, float]]:
        """Returns most suitable candidate for fixing incorrect words

            Args:
                current_word: The words that need correction
                correct_words_before: correct words before current word
                words_after: not processed yet words after current word
                candidates: Candidates for right word

            Returns:
                Most suitable candidate and it's score
        """

    @abstractmethod
    def pick_top_candidate_with_next_word(self, current_word: Word,
                                          next_word: Word,
                                          correct_words_before: List[Word],
                                          words_after: List[Word],
                                          candidates: List[CandidateWord]) -> Optional[Tuple[CandidateWord, float]]:
        """Returns most suitable candidate (with next word) for fixing incorrect words

            Args:
                current_word: The words that need correction
                next_word: Next word
                correct_words_before: correct words before current word
                words_after: not processed yet words after current word
                candidates: Candidates for right word

            Returns:
                Most suitable candidate and it's score
        """
